'''
The function TestAllModels(n) tries to compute the structure of End(Jac(C))
for all hyperelliptic curves C : y^2= x^5 + a4 x^4 + ... + a0
where the a_i are integers, |a_i| <= n
'''

attach("Irreducibility.sage");
attach("Reducibility.sage");
attach("CM.sage");
attach("RM.sage");
attach("GeometricallyTrivial.sage");

def PrintOutput(bound, string,suffix=""):
  f1=open('./Output'+suffix+str(bound), 'a')
  f1.write(string + "\n");
  print(string);
  f1.close();



def TestAllModels(bound):
  f1=open('./Output'+str(bound), 'a')
  f1.write("");
  f1.close()
  os.remove('./Output'+str(bound));
  maximP=0
  counter=0
  singular=0
  irreducible=0
  reducible=[0 for j in range(8)]
  endZ=0
  CM=0
  QM=0
  RM=[0 for j in range(30)]
  
  for a0 in range(-bound,bound+1):
    for a1 in range(-bound,bound+1):
      for a2 in range(-bound,bound+1):
	  for a3 in range(-bound,bound+1):
	      for a4 in range(0,bound+1):
		counter=counter+1
		poly=ZZ['x']([a0,a1,a2,a3,a4,1])
		discPoly=poly.discriminant();
		if (discPoly <> 0) : 
			isProvenIrreducible=false;
			res=test_certifyGeometricallyIrreducible(a0,a1,a2,a3,a4);
			if(res>maximP) :
				maximP=res;
			if (res>=0) :
				irreducible=irreducible+1;
				'''
                                The Jacobian is geometrically irreducible
				'''
				isProvenIrreducible=true;
			if isProvenIrreducible :
				if(test_EndZ_Hall(a0,a1,a2,a3,a4)==false) :
					if (test_EndZ_Zarhin(a0,a1,a2,a3,a4) == false) :
						  if CertifyEndZ(a0,a1,a2,a3,a4,1,0):
						  	endZ=endZ+1;
						  else:
				                          if test_CertifyCM(a0,a1,a2,a3,a4,1,0,false):
							    CM=CM+1;
							    PrintOutput(bound, "CM curve with defining equation y^2 = " + str(poly) + "." );
							  else :
							    RMDiscriminant=CertifyRM(a0,a1,a2,a3,a4,1,0,false);
							    if RMDiscriminant>0 :
							      RM[RMDiscriminant]=RM[RMDiscriminant]+1;
							      PrintOutput(bound, "RM curve with defining equation y^2 = " + str(poly) + str(". The discriminant of the endomorphism ring is ")+str(RMDiscriminant) + "." );
							    else :
							      PrintOutput(bound, "Found a (likely) RM curve with uncertain endomorphism ring. Defining polynomial: " + str(poly) );
					else:
						'''
						passed Zarhin's test, hence GeometricEnd=Z
						'''
						endZ=endZ+1;
				else:
					'''
					passed Hall's test, hence GeometricEnd=Z
					'''
					endZ=endZ+1;
			else:
				  '''
				  if the Jacobian is not ProvenIrreducible, test for small reducibility or QM
			          '''
				  SplittingOrder=test_Reducibility(a0,a1,a2,a3,a4);
				  if SplittingOrder>0 :
				    reducible[SplittingOrder]=reducible[SplittingOrder]+1;
				    PrintOutput(bound, "("+str(SplittingOrder)+","+str(SplittingOrder)+")-split Jacobian. Defining equation y^2 = " + str(poly) );
				  else :
				    if(test_certifyNonQM(a0,a1,a2,a3,a4)==false):
					    QMResult=test_certifyQM(a0,a1,a2,a3,a4);
					    if(QMResult!=false ):
						PrintOutput (bound, "QM curve with defining equation y^2 = " + str(poly) + str(". Quaternion ring corresponding to the matrix:\n") + str(QMResult) );
						QM=QM+1;
						irreducible=irreducible+1;
					    else:
						PrintOutput (bound, "Found a (likely) QM curve. Defining polynomial: " + str(poly) );
				    else:
						PrintOutput (bound, "Found a (likely) split Jacobian, with order of splitting at least 6. Defining polynomial: " + str(poly) );
		else :
			singular=singular+1;


  PrintOutput(bound, "Models tested: " + str(counter));
  PrintOutput(bound, "Singular: " + str(singular));
  PrintOutput(bound, "Irreducible: " + str(irreducible));
  PrintOutput(bound, "Geometric Endomorphism = Z: " + str(endZ));
  PrintOutput(bound, "Geometric CM: " + str(CM));
  PrintOutput(bound, "Geometric RM: " + str(RM));
  PrintOutput(bound, "Geometric QM: " + str(QM));
  PrintOutput(bound, "Reducible: " + str(reducible));
  PrintOutput(bound, "Cannot determine reducibility/irreducibility: " + str(counter-singular-irreducible-sum(reducible)));
  PrintOutput(bound, "Cannot determine endomorphism ring: " + str(counter-singular-endZ-sum(reducible)-sum(RM)-CM-QM));
  PrintOutput(bound, "Greatest prime used to prove irreducibility: " + str(maximP));
